@@ -18,14 +18,14 @@ typedef enum {
    /* Misc */
    ATOM_COMPOUND_TEXT,
    ATOM_UTF8_STRING,
-   ATOM_XSETROOT_ID,
+   ATOM_XROOTPMAP_ID,
+   ATOM_MANAGER,
 
    /* Standard atoms */
    ATOM_WM_STATE,
    ATOM_WM_PROTOCOLS,
    ATOM_WM_DELETE_WINDOW,
    ATOM_WM_TAKE_FOCUS,
-   ATOM_WM_LOCALE_NAME,
    ATOM_WM_CHANGE_STATE,
    ATOM_WM_COLORMAP_WINDOWS,
 
@@ -39,6 +39,7 @@ typedef enum {
    ATOM_NET_ACTIVE_WINDOW,
    ATOM_NET_WORKAREA,
    ATOM_NET_SUPPORTING_WM_CHECK,
+   ATOM_NET_SHOWING_DESKTOP,
    ATOM_NET_FRAME_EXTENTS,
    ATOM_NET_WM_DESKTOP,
 
@@ -48,6 +49,13 @@ typedef enum {
    ATOM_NET_WM_STATE_MAXIMIZED_HORZ,
    ATOM_NET_WM_STATE_SHADED,
    ATOM_NET_WM_STATE_FULLSCREEN,
+   ATOM_NET_WM_STATE_HIDDEN,
+   ATOM_NET_WM_STATE_SKIP_TASKBAR,
+   ATOM_NET_WM_STATE_SKIP_PAGER,
+   ATOM_NET_WM_STATE_BELOW,
+   ATOM_NET_WM_STATE_ABOVE,
+   ATOM_NET_WM_STATE_DEMANDS_ATTENTION,
+   ATOM_NET_WM_STATE_FOCUSED,
 
    ATOM_NET_WM_ALLOWED_ACTIONS,
    ATOM_NET_WM_ACTION_MOVE,
@@ -55,35 +63,49 @@ typedef enum {
    ATOM_NET_WM_ACTION_MINIMIZE,
    ATOM_NET_WM_ACTION_SHADE,
    ATOM_NET_WM_ACTION_STICK,
+   ATOM_NET_WM_ACTION_FULLSCREEN,
    ATOM_NET_WM_ACTION_MAXIMIZE_HORZ,
    ATOM_NET_WM_ACTION_MAXIMIZE_VERT,
    ATOM_NET_WM_ACTION_CHANGE_DESKTOP,
    ATOM_NET_WM_ACTION_CLOSE,
+   ATOM_NET_WM_ACTION_BELOW,
+   ATOM_NET_WM_ACTION_ABOVE,
 
    ATOM_NET_CLOSE_WINDOW,
    ATOM_NET_MOVERESIZE_WINDOW,
+   ATOM_NET_RESTACK_WINDOW,
+   ATOM_NET_REQUEST_FRAME_EXTENTS,
 
+   ATOM_NET_WM_PID,
    ATOM_NET_WM_NAME,
+   ATOM_NET_WM_VISIBLE_NAME,
+   ATOM_NET_WM_HANDLED_ICONS,
    ATOM_NET_WM_ICON,
+   ATOM_NET_WM_ICON_NAME,
+   ATOM_NET_WM_USER_TIME,
+   ATOM_NET_WM_USER_TIME_WINDOW,
+   ATOM_NET_WM_VISIBLE_ICON_NAME,
    ATOM_NET_WM_WINDOW_TYPE,
    ATOM_NET_WM_WINDOW_TYPE_DESKTOP,
    ATOM_NET_WM_WINDOW_TYPE_DOCK,
+   ATOM_NET_WM_WINDOW_TYPE_SPLASH,
+   ATOM_NET_WM_WINDOW_TYPE_DIALOG,
+   ATOM_NET_WM_WINDOW_TYPE_NORMAL,
+   ATOM_NET_WM_WINDOW_TYPE_MENU,
+   ATOM_NET_WM_WINDOW_TYPE_NOTIFICATION,
+   ATOM_NET_WM_WINDOW_TYPE_TOOLBAR,
+   ATOM_NET_WM_WINDOW_TYPE_UTILITY,
 
    ATOM_NET_CLIENT_LIST,
    ATOM_NET_CLIENT_LIST_STACKING,
 
    ATOM_NET_WM_STRUT_PARTIAL,
+   ATOM_NET_WM_WINDOW_OPACITY,
    ATOM_NET_WM_STRUT,
+   ATOM_NET_WM_MOVERESIZE,
 
    ATOM_NET_SYSTEM_TRAY_OPCODE,
-
-   /* GNOME atoms */
-   ATOM_WIN_LAYER,
-   ATOM_WIN_STATE,
-   ATOM_WIN_WORKSPACE_COUNT,
-   ATOM_WIN_WORKSPACE,
-   ATOM_WIN_SUPPORTING_WM_CHECK,
-   ATOM_WIN_PROTOCOLS,
+   ATOM_NET_SYSTEM_TRAY_ORIENTATION,
 
    /* MWM atoms */
    ATOM_MOTIF_WM_HINTS,
@@ -91,80 +113,91 @@ typedef enum {
    /* JWM-specific atoms. */
    ATOM_JWM_RESTART,
    ATOM_JWM_EXIT,
+   ATOM_JWM_RELOAD,
+   ATOM_JWM_WM_STATE_MAXIMIZED_TOP,
+   ATOM_JWM_WM_STATE_MAXIMIZED_BOTTOM,
+   ATOM_JWM_WM_STATE_MAXIMIZED_LEFT,
+   ATOM_JWM_WM_STATE_MAXIMIZED_RIGHT,
 
    ATOM_COUNT
 } AtomType;
 
-#define FIRST_NET_ATOM ATOM_NET_SUPPORTED
-#define LAST_NET_ATOM  ATOM_NET_SYSTEM_TRAY_OPCODE
+extern const char jwmRestart[];
+extern const char jwmExit[];
+extern const char jwmReload[];
+extern const char managerProperty[];
 
-#define FIRST_WIN_ATOM ATOM_WIN_LAYER
-#define LAST_WIN_ATOM  ATOM_WIN_PROTOCOLS
+#define FIRST_NET_ATOM ATOM_NET_SUPPORTED
+#define LAST_NET_ATOM  ATOM_NET_WM_STRUT
 
 #define FIRST_MWM_ATOM ATOM_MOTIF_WM_HINTS
 #define LAST_MWM_ATOM  ATOM_MOTIF_WM_HINTS
 
-#define WIN_STATE_STICKY          (1UL << 0)
-#define WIN_STATE_MINIMIZED       (1UL << 1)
-#define WIN_STATE_MAXIMIZED_VERT  (1UL << 2)
-#define WIN_STATE_MAXIMIZED_HORIZ (1UL << 3)
-#define WIN_STATE_HIDDEN          (1UL << 4)
-#define WIN_STATE_SHADED          (1UL << 5)
-#define WIN_STATE_HIDE_WORKSPACE  (1UL << 6)
-#define WIN_STATE_HIDE_TRANSIENT  (1UL << 7)
-#define WIN_STATE_FIXED_POSITION  (1UL << 8)
-#define WIN_STATE_ARRANGE_IGNORE  (1UL << 9)
-
-#define WIN_HINT_SKIP_FOCUS      (1UL << 0)
-#define WIN_HINT_SKIP_WINLIST    (1UL << 1)
-#define WIN_HINT_SKIP_TASKBAR    (1UL << 2)
-#define WIN_HINT_GROUP_TRANSIENT (1UL << 3)
-#define WIN_HINT_FOCUS_ON_CLICK  (1UL << 4)
-
 /** Enumeration of window layers. */
-typedef enum {
-   LAYER_BOTTOM              = 0,
-   LAYER_NORMAL              = 4,
-   DEFAULT_TRAY_LAYER        = 8,
-   LAYER_TOP                 = 12,
-   LAYER_COUNT               = 13
-} WinLayerType;
+typedef unsigned char WinLayerType;
+#define LAYER_DESKTOP   0
+#define LAYER_BELOW     1
+#define LAYER_NORMAL    2
+#define LAYER_ABOVE     3
+#define LAYER_COUNT     4
+
+#define FIRST_LAYER        LAYER_DESKTOP
+#define LAST_LAYER         LAYER_ABOVE
+#define DEFAULT_TRAY_LAYER LAYER_ABOVE
+
+/** Enumeration of window type. */
+typedef unsigned char WindowType;
+#define WINDOW_TYPE_DESKTOP        0
+#define WINDOW_TYPE_DOCK           1
+#define WINDOW_TYPE_SPLASH         2
+#define WINDOW_TYPE_DIALOG         3
+#define WINDOW_TYPE_NORMAL         4
+#define WINDOW_TYPE_MENU           5
+#define WINDOW_TYPE_NOTIFICATION   6
+#define WINDOW_TYPE_TOOLBAR        7
+#define WINDOW_TYPE_UTILITY        8
+#define WINDOW_TYPE_COUNT          9
 
 /** Client state information. */
 typedef struct ClientState {
-   unsigned int status;    /**< Status bit mask. */
-   unsigned int border;    /**< Border bit mask. */
-   unsigned int layer;     /**< Window layer. */
-   unsigned int desktop;   /**< Desktop. */
+   unsigned int status;          /**< Status bit mask. */
+   unsigned int opacity;         /**< Opacity (0 - 0xFFFFFFFF). */
+   unsigned short border;        /**< Border bit mask. */
+   unsigned short desktop;       /**< Desktop. */
+   unsigned char maxFlags;       /**< Maximization status. */
+   unsigned char layer;          /**< Current window layer. */
+   unsigned char defaultLayer;   /**< Default window layer. */
+   unsigned char windowType;     /**< Window type. */
 } ClientState;
 
-/** Client protocols (to be used as a bit mask). */
-typedef enum {
-   PROT_NONE       = 0,
-   PROT_DELETE     = 1,
-   PROT_TAKE_FOCUS = 2
-} ClientProtocolType;
-
 extern Atom atoms[ATOM_COUNT];
 
-void InitializeHints();
-void StartupHints();
-void ShutdownHints();
-void DestroyHints();
+/*@{*/
+#define InitializeHints()  (void)(0)
+void StartupHints(void);
+#define ShutdownHints()    (void)(0)
+#define DestroyHints()     (void)(0)
+/*@}*/
 
 /** Determine the current desktop. */
-void ReadCurrentDesktop();
+void ReadCurrentDesktop(void);
 
-/** Read client protocols
+/** Read client info.
  * @param np The client.
+ * @param alreadyMapped Set if the client is already mapped.
  */
-void ReadClientProtocols(struct ClientNode *np);
+void ReadClientInfo(struct ClientNode *np, char alreadyMapped);
 
 /** Read a client's name.
  * @param np The client.
  */
 void ReadWMName(struct ClientNode *np);
 
+/** Read a client's machine.
+ * @param np The client.
+ */
+void ReadWMMachine(struct ClientNode *np);
+
 /** Read a client's class.
  * @param np The client.
  */
@@ -177,9 +210,9 @@ void ReadWMNormalHints(struct ClientNode *np);
 
 /** Read the WM_PROTOCOLS property for a window.
  * @param w The window.
- * @return The protocols.
+ * @param state The client state to update.
  */
-ClientProtocolType ReadWMProtocols(Window w);
+void ReadWMProtocols(Window w, ClientState *state);
 
 /** Read colormap information for a client.
  * @param np The client.
@@ -193,30 +226,55 @@ void ReadWinLayer(struct ClientNode *np);
 
 /** Read the current state of a window.
  * @param win The window.
+ * @param alreadyMapped Set if the window is already mapped.
  * @return The window state.
  */
-ClientState ReadWindowState(Window win);
+ClientState ReadWindowState(Window win, char alreadyMapped);
+
+/** Read WM hints.
+ * @param win The window.
+ * @param state The state hints to update.
+ * @param alreadyMapped Set if the window is already mapped.
+ */
+void ReadWMHints(Window win, ClientState *state, char alreadyMapped);
+
+/** Read opacity.
+ * @param win The window.
+ * @param opacity The opacity to update.
+ */
+void ReadWMOpacity(Window win, unsigned *opacity);
 
 /** Set the state of a client window.
+ * Note that this will call WriteNetState.
  * @param np The client.
  */
 void WriteState(struct ClientNode *np);
 
+/** Set _NET_WM_STATE.
+ * @param np The client.
+ */
+void WriteNetState(struct ClientNode *np);
+
+/** Set the opacity of a client window.
+ * @param np The client.
+ * @param opacity The opacity to set.
+ * @param force Set the opacity even if it hasn't changed.
+ */
+void SetOpacity(struct ClientNode *np, unsigned int opacity, char force);
+
+/** Set the frame extents of a window.
+ * @param win The window.
+ * @param state The client state.
+ */
+void WriteFrameExtents(Window win, const ClientState *state);
+
 /** Read a cardinal atom.
  * @param window The window.
  * @param atom The atom to read.
  * @param value A pointer to the location to save the atom.
  * @return 1 on success, 0 on failure.
  */
-int GetCardinalAtom(Window window, AtomType atom, unsigned long *value);
-
-/** Read a window atom.
- * @param window The window.
- * @param atom The atom to read.
- * @param value A pointer to the location to save the atom.
- * @return 1 on success, 0 on failure.
- */
-int GetWindowAtom(Window window, AtomType atom, Window *value);
+char GetCardinalAtom(Window window, AtomType atom, unsigned long *value);
 
 /** Set a cardinal atom.
  * @param window The window.
@@ -225,6 +283,14 @@ int GetWindowAtom(Window window, AtomType atom, Window *value);
  */
 void SetCardinalAtom(Window window, AtomType atom, unsigned long value);
 
+/** Read a window atom.
+ * @param window The window.
+ * @param atom The atom to read.
+ * @param value A pointer to the location to save the atom.
+ * @return 1 on success, 0 on failure.
+ */
+char GetWindowAtom(Window window, AtomType atom, Window *value);
+
 /** Set a window atom.
  * @param window The window.
  * @param atom The atom to set.
@@ -239,5 +305,12 @@ void SetWindowAtom(Window window, AtomType atom, unsigned long value);
  */
 void SetPixmapAtom(Window window, AtomType atom, Pixmap value);
 
+/** Set an atom atom.
+ * @param window The window.
+ * @param atom The atom to set.
+ * @param value The value.
+ */
+void SetAtomAtom(Window window, AtomType atom, AtomType value);
+
 #endif /* HINT_H */
 
