@@ -13,19 +13,20 @@
 struct TrayComponentType;
 
 /*@{*/
-void InitializePager();
-void StartupPager();
-void ShutdownPager();
-void DestroyPager();
+#define InitializePager()  (void)(0)
+#define StartupPager()     (void)(0)
+void ShutdownPager(void);
+void DestroyPager(void);
 /*@}*/
 
 /** Create a pager tray component.
+ * @param labeled Set to label the pager.
  * @return A new pager tray component.
  */
-struct TrayComponentType *CreatePager();
+struct TrayComponentType *CreatePager(char labeled);
 
 /** Update pagers. */
-void UpdatePager();
+void UpdatePager(void);
 
 #endif /* PAGER_H */
 
