@@ -16,13 +16,20 @@
 #include "error.h"
 #include "taskbar.h"
 #include "menu.h"
-#include "timing.h"
+#include "screen.h"
+#include "settings.h"
+#include "event.h"
+#include "client.h"
+#include "misc.h"
+#include "hint.h"
 
 #define DEFAULT_TRAY_WIDTH 32
 #define DEFAULT_TRAY_HEIGHT 32
 
+#define TRAY_BORDER_SIZE   1
+
 static TrayType *trays;
-static Window supportingWindow;
+static unsigned int trayCount;
 
 static void HandleTrayExpose(TrayType *tp, const XExposeEvent *event);
 static void HandleTrayEnterNotify(TrayType *tp, const XCrossingEvent *event);
@@ -32,25 +39,30 @@ static void HandleTrayButtonPress(TrayType *tp, const XButtonEvent *event);
 static void HandleTrayButtonRelease(TrayType *tp, const XButtonEvent *event);
 static void HandleTrayMotionNotify(TrayType *tp, const XMotionEvent *event);
 
-static void ComputeTraySize(TrayType *tp);
+static void ComputeTrayGeometry(TrayType *tp);
 static int ComputeMaxWidth(TrayType *tp);
 static int ComputeTotalWidth(TrayType *tp);
 static int ComputeMaxHeight(TrayType *tp);
 static int ComputeTotalHeight(TrayType *tp);
-static int CheckHorizontalFill(TrayType *tp);
-static int CheckVerticalFill(TrayType *tp);
+static char CheckHorizontalFill(TrayType *tp);
+static char CheckVerticalFill(TrayType *tp);
 static void LayoutTray(TrayType *tp, int *variableSize,
-   int *variableRemainder);
+                       int *variableRemainder);
+
+static void SignalTray(const TimeType *now, int x, int y, Window w,
+                       void *data);
+
 
 /** Initialize tray data. */
-void InitializeTray() {
+void InitializeTray(void)
+{
    trays = NULL;
-   supportingWindow = None;
+   trayCount = 0;
 }
 
 /** Startup trays. */
-void StartupTray() {
-
+void StartupTray(void)
+{
    XSetWindowAttributes attr;
    unsigned long attrMask;
    TrayType *tp;
@@ -83,43 +95,53 @@ void StartupTray() {
          | PointerMotionMask;
 
       attrMask |= CWBackPixel;
-      attr.background_pixel = colors[COLOR_TRAY_BG];
+      attr.background_pixel = colors[COLOR_TRAY_BG2];
 
+      Assert(tp->width > 0);
+      Assert(tp->height > 0);
       tp->window = JXCreateWindow(display, rootWindow,
-         tp->x, tp->y, tp->width, tp->height,
-         0, rootDepth, InputOutput, rootVisual, attrMask, &attr);
+                                  tp->x, tp->y, tp->width, tp->height, 0,
+                                  rootDepth, InputOutput,
+                                  rootVisual, attrMask, &attr);
+      SetAtomAtom(tp->window, ATOM_NET_WM_WINDOW_TYPE,
+                  ATOM_NET_WM_WINDOW_TYPE_DOCK);
+
+      if(settings.trayOpacity < UINT_MAX) {
+         SetCardinalAtom(tp->window, ATOM_NET_WM_WINDOW_OPACITY,
+                         settings.trayOpacity);
+      }
 
       SetDefaultCursor(tp->window);
 
       /* Create and layout items on the tray. */
-      xoffset = tp->border;
-      yoffset = tp->border;
+      xoffset = TRAY_BORDER_SIZE;
+      yoffset = TRAY_BORDER_SIZE;
       for(cp = tp->components; cp; cp = cp->next) {
 
          if(cp->Create) {
             if(tp->layout == LAYOUT_HORIZONTAL) {
-               height = tp->height - 2 * tp->border;
+               height = tp->height - TRAY_BORDER_SIZE * 2;
                width = cp->width;
                if(width == 0) {
                   width = variableSize;
                   if(variableRemainder) {
-                     ++width;
-                     --variableRemainder;
+                     width += 1;
+                     variableRemainder -= 1;
                   }
                }
             } else {
-               width = tp->width - 2 * tp->border;
+               width = tp->width - TRAY_BORDER_SIZE * 2;
                height = cp->height;
                if(height == 0) {
                   height = variableSize;
                   if(variableRemainder) {
-                     ++height;
-                     --variableRemainder;
+                     height += 1;
+                     variableRemainder -= 1;
                   }
                }
             }
-            cp->width = width;
-            cp->height = height;
+            cp->width = Max(1, width);
+            cp->height = Max(1, height);
             (cp->Create)(cp);
          }
 
@@ -130,7 +152,7 @@ void StartupTray() {
 
          if(cp->window != None) {
             JXReparentWindow(display, cp->window, tp->window,
-               xoffset, yoffset);
+                             xoffset, yoffset);
          }
 
          if(tp->layout == LAYOUT_HORIZONTAL) {
@@ -143,16 +165,17 @@ void StartupTray() {
       /* Show the tray. */
       JXMapWindow(display, tp->window);
 
+      trayCount += 1;
+
    }
 
-   UpdatePager();
-   UpdateTaskBar();
-
+   RequirePagerUpdate();
+   RequireTaskUpdate();
 }
 
 /** Shutdown trays. */
-void ShutdownTray() {
-
+void ShutdownTray(void)
+{
    TrayType *tp;
    TrayComponentType *cp;
 
@@ -164,23 +187,19 @@ void ShutdownTray() {
       }
       JXDestroyWindow(display, tp->window);
    }
-
-   if(supportingWindow != None) {
-      XDestroyWindow(display, supportingWindow);
-      supportingWindow = None;
-   }
-
 }
 
 /** Destroy tray data. */
-void DestroyTray() {
-
+void DestroyTray(void)
+{
    TrayType *tp;
    TrayComponentType *cp;
 
    while(trays) {
       tp = trays->next;
-
+      if(trays->autoHide != THIDE_OFF) {
+         UnregisterCallback(SignalTray, trays);
+      }
       while(trays->components) {
          cp = trays->components->next;
          Release(trays->components);
@@ -190,29 +209,32 @@ void DestroyTray() {
 
       trays = tp;
    }
-
 }
 
 /** Create an empty tray. */
-TrayType *CreateTray() {
-
+TrayType *CreateTray(void)
+{
    TrayType *tp;
 
    tp = Allocate(sizeof(TrayType));
 
+   tp->requestedX = 0;
+   tp->requestedY = -1;
+   tp->screen = 0;
    tp->x = 0;
    tp->y = -1;
    tp->requestedWidth = 0;
    tp->requestedHeight = 0;
    tp->width = 0;
    tp->height = 0;
-   tp->border = 1;
    tp->layer = DEFAULT_TRAY_LAYER;
    tp->layout = LAYOUT_HORIZONTAL;
    tp->valign = TALIGN_FIXED;
    tp->halign = TALIGN_FIXED;
 
-   tp->autoHide = 0;
+   GetCurrentTime(&tp->showTime);
+   tp->autoHide = THIDE_OFF;
+   tp->autoHideDelay = 0;
    tp->hidden = 0;
 
    tp->window = None;
@@ -224,12 +246,11 @@ TrayType *CreateTray() {
    trays = tp;
 
    return tp;
-
 }
 
 /** Create an empty tray component. */
-TrayComponentType *CreateTrayComponent() {
-
+TrayComponentType *CreateTrayComponent(void)
+{
    TrayComponentType *cp;
 
    cp = Allocate(sizeof(TrayComponentType));
@@ -257,21 +278,17 @@ TrayComponentType *CreateTrayComponent() {
    cp->ProcessButtonPress = NULL;
    cp->ProcessButtonRelease = NULL;
    cp->ProcessMotionEvent = NULL;
+   cp->Redraw = NULL;
 
    cp->next = NULL;
 
    return cp;
-
 }
 
 /** Add a tray component to a tray. */
-void AddTrayComponent(TrayType *tp, TrayComponentType *cp) {
-
-   Assert(tp);
-   Assert(cp);
-
+void AddTrayComponent(TrayType *tp, TrayComponentType *cp)
+{
    cp->tray = tp;
-
    if(tp->componentsTail) {
       tp->componentsTail->next = cp;
    } else {
@@ -279,12 +296,11 @@ void AddTrayComponent(TrayType *tp, TrayComponentType *cp) {
    }
    tp->componentsTail = cp;
    cp->next = NULL;
-
 }
 
 /** Compute the max component width. */
-int ComputeMaxWidth(TrayType *tp) {
-
+int ComputeMaxWidth(TrayType *tp)
+{
    TrayComponentType *cp;
    int result;
    int temp;
@@ -293,7 +309,6 @@ int ComputeMaxWidth(TrayType *tp) {
    for(cp = tp->components; cp; cp = cp->next) {
       temp = cp->width;
       if(temp > 0) {
-         temp += 2 * tp->border;
          if(temp > result) {
             result = temp;
          }
@@ -301,27 +316,25 @@ int ComputeMaxWidth(TrayType *tp) {
    }
 
    return result;
-
 }
 
 /** Compute the total width of a tray. */
-int ComputeTotalWidth(TrayType *tp) {
-
+int ComputeTotalWidth(TrayType *tp)
+{
    TrayComponentType *cp;
    int result;
 
-   result = 2 * tp->border;
+   result = 2 * TRAY_BORDER_SIZE;
    for(cp = tp->components; cp; cp = cp->next) {
       result += cp->width;
    }
 
    return result;
-
 }
 
 /** Compute the max component height. */
-int ComputeMaxHeight(TrayType *tp) {
-
+int ComputeMaxHeight(TrayType *tp)
+{
    TrayComponentType *cp;
    int result;
    int temp;
@@ -330,7 +343,6 @@ int ComputeMaxHeight(TrayType *tp) {
    for(cp = tp->components; cp; cp = cp->next) {
       temp = cp->height;
       if(temp > 0) {
-         temp += 2 * tp->border;
          if(temp > result) {
             result = temp;
          }
@@ -338,27 +350,25 @@ int ComputeMaxHeight(TrayType *tp) {
    }
 
    return result;
-
 }
 
 /** Compute the total height of a tray. */
-int ComputeTotalHeight(TrayType *tp) {
-
+int ComputeTotalHeight(TrayType *tp)
+{
    TrayComponentType *cp;
    int result;
 
-   result = 2 * tp->border;
+   result = 2 * TRAY_BORDER_SIZE;
    for(cp = tp->components; cp; cp = cp->next) {
       result += cp->height;
    }
 
    return result;
-
 }
 
 /** Check if the tray fills the screen horizontally. */
-int CheckHorizontalFill(TrayType *tp) {
-
+char CheckHorizontalFill(TrayType *tp)
+{
    TrayComponentType *cp;
 
    for(cp = tp->components; cp; cp = cp->next) {
@@ -368,12 +378,11 @@ int CheckHorizontalFill(TrayType *tp) {
    }
 
    return 0;
-
 }
 
 /** Check if the tray fills the screen vertically. */
-int CheckVerticalFill(TrayType *tp) {
-
+char CheckVerticalFill(TrayType *tp)
+{
    TrayComponentType *cp;
 
    for(cp = tp->components; cp; cp = cp->next) {
@@ -383,53 +392,93 @@ int CheckVerticalFill(TrayType *tp) {
    }
 
    return 0;
-
 }
 
-/** Compute the size of a tray. */
-void ComputeTraySize(TrayType *tp) {
-
+/** Compute the size and position of a tray. */
+void ComputeTrayGeometry(TrayType *tp)
+{
+   const ScreenType *sp = GetScreen(tp->screen);
    TrayComponentType *cp;
+   int x, y;
+
+   /* Set requested position.
+    * The requested coordinates (if provided) are screen-relative.
+    * This will compute x and y to be in root-window coordinates.
+    */
+   x = 0;
+   y = 0;
+   if(tp->halign == TALIGN_FIXED) {
+      x = tp->requestedX;
+      if(x < 0) {
+         x += sp->width;
+      }
+   }
+   if(tp->valign == TALIGN_FIXED) {
+      y = tp->requestedY;
+      if(y < 0) {
+         y += sp->height;
+      }
+   }
+   x += sp->x;
+   y += sp->y;
+
+   /* Set requested sizes. */
+   if(tp->requestedWidth >= 0) {
+      tp->width = tp->requestedWidth;
+   } else {
+      tp->width = sp->width + tp->requestedWidth - (x - sp->x);
+   }
+   if(tp->requestedHeight >= 0) {
+      tp->height = tp->requestedHeight;
+   } else {
+      tp->height = sp->height + tp->requestedHeight - (y - sp->y);
+   }
+   for(cp = tp->components; cp; cp = cp->next) {
+      if(cp->requestedWidth) {
+         cp->width = cp->requestedWidth;
+      } else {
+         cp->width = 0;
+      }
+      if(cp->requestedHeight) {
+         cp->height = cp->requestedHeight;
+      } else {
+         cp->height = 0;
+      }
+   }
 
    /* Determine the first dimension. */
    if(tp->layout == LAYOUT_HORIZONTAL) {
-
-      if(tp->height == 0) {
-         tp->height = ComputeMaxHeight(tp);
+      if(tp->requestedHeight == 0) {
+         tp->height = ComputeMaxHeight(tp) + TRAY_BORDER_SIZE * 2;
       }
-
       if(tp->height == 0) {
          tp->height = DEFAULT_TRAY_HEIGHT;
       }
-
    } else {
-
-      if(tp->width == 0) {
-         tp->width = ComputeMaxWidth(tp);
+      if(tp->requestedWidth == 0) {
+         tp->width = ComputeMaxWidth(tp) + TRAY_BORDER_SIZE * 2;
       }
-
       if(tp->width == 0) {
          tp->width = DEFAULT_TRAY_WIDTH;
       }
-
    }
 
    /* Now at least one size is known. Inform the components. */
    for(cp = tp->components; cp; cp = cp->next) {
       if(cp->SetSize) {
          if(tp->layout == LAYOUT_HORIZONTAL) {
-            (cp->SetSize)(cp, 0, tp->height - 2 * tp->border);
+            (cp->SetSize)(cp, 0, tp->height - TRAY_BORDER_SIZE * 2);
          } else {
-            (cp->SetSize)(cp, tp->width - 2 * tp->border, 0);
+            (cp->SetSize)(cp, tp->width - TRAY_BORDER_SIZE * 2, 0);
          }
       }
    }
 
    /* Determine the missing dimension. */
    if(tp->layout == LAYOUT_HORIZONTAL) {
-      if(tp->width == 0) {
+      if(tp->requestedWidth == 0) {
          if(CheckHorizontalFill(tp)) {
-            tp->width = rootWidth;
+            tp->width = sp->width - abs(tp->requestedX);
          } else {
             tp->width = ComputeTotalWidth(tp);
          }
@@ -438,9 +487,9 @@ void ComputeTraySize(TrayType *tp) {
          }
       }
    } else {
-      if(tp->height == 0) {
+      if(tp->requestedHeight == 0) {
          if(CheckVerticalFill(tp)) {
-            tp->height = rootHeight;
+            tp->height = sp->height - abs(tp->requestedY);
          } else {
             tp->height = ComputeTotalHeight(tp);
          }
@@ -453,43 +502,46 @@ void ComputeTraySize(TrayType *tp) {
    /* Compute the tray location. */
    switch(tp->valign) {
    case TALIGN_TOP:
-      tp->y = 0;
+      tp->y = sp->y;
       break;
    case TALIGN_BOTTOM:
-      tp->y = rootHeight - tp->height + 1;
+      tp->y = sp->y + sp->height - tp->height;
       break;
    case TALIGN_CENTER:
-      tp->y = rootHeight / 2 - tp->height / 2;
+      tp->y = sp->y + (sp->height - tp->height) / 2;
       break;
    default:
-      if(tp->y < 0) {
-         tp->y = rootHeight + tp->y - tp->height + 1;
+      if(tp->requestedY < 0) {
+         tp->y = y - tp->height;
+      } else {
+         tp->y = y;
       }
       break;
    }
 
    switch(tp->halign) {
    case TALIGN_LEFT:
-      tp->x = 0;
+      tp->x = sp->x;
       break;
    case TALIGN_RIGHT:
-      tp->x = rootWidth - tp->width + 1;
+      tp->x = sp->x + sp->width - tp->width;
       break;
    case TALIGN_CENTER:
-      tp->x = rootWidth / 2 - tp->width / 2;
+      tp->x = sp->x + (sp->width - tp->width) / 2;
       break;
    default:
-      if(tp->x < 0) {
-         tp->x = rootWidth + tp->x - tp->width + 1;
+      if(tp->requestedX < 0) {
+         tp->x = x - tp->width;
+      } else {
+         tp->x = x;
       }
       break;
    }
-
 }
 
 /** Display a tray (for autohide). */
-void ShowTray(TrayType *tp) {
-
+void ShowTray(TrayType *tp)
+{
    Window win1, win2;
    int winx, winy;
    unsigned int mask;
@@ -498,54 +550,99 @@ void ShowTray(TrayType *tp) {
    if(tp->hidden) {
 
       tp->hidden = 0;
+      GetCurrentTime(&tp->showTime);
       JXMoveWindow(display, tp->window, tp->x, tp->y);
 
       JXQueryPointer(display, rootWindow, &win1, &win2,
-         &mousex, &mousey, &winx, &winy, &mask);
-      SetMousePosition(mousex, mousey);
+                     &mousex, &mousey, &winx, &winy, &mask);
+      SetMousePosition(mousex, mousey, win2);
 
    }
+}
 
+/** Show all trays. */
+void ShowAllTrays(void)
+{
+   TrayType *tp;
+
+   if(shouldExit) {
+      return;
+   }
+
+   for(tp = trays; tp; tp = tp->next) {
+      ShowTray(tp);
+   }
 }
 
 /** Hide a tray (for autohide). */
-void HideTray(TrayType *tp) {
-
+void HideTray(TrayType *tp)
+{
+   const ScreenType *sp;
+   LayoutType layout;
    int x, y;
 
+   /* Don't hide if the tray is raised. */
+   if(tp->autoHide & THIDE_RAISED) {
+      return;
+   }
+
    tp->hidden = 1;
 
-   /* Determine where to move the tray. */
-   if(tp->layout == LAYOUT_HORIZONTAL) {
-
-      x = tp->x;
-
-      if(tp->y >= rootHeight / 2) {
-         y = rootHeight - 1;
+   /* Derive the location for hiding the tray. */
+   sp = GetCurrentScreen(tp->x, tp->y);
+   layout = tp->autoHide;
+   if(layout == THIDE_ON) {
+      if(tp->layout == LAYOUT_HORIZONTAL) {
+         if(tp->y >= sp->height / 2) {
+            layout = THIDE_BOTTOM;
+         } else {
+            layout = THIDE_TOP;
+         }
       } else {
-         y = 1 - tp->height;
+         if(tp->x >= sp->width / 2) {
+            layout = THIDE_RIGHT;
+         } else {
+            layout = THIDE_LEFT;
+         }
       }
-
-   } else {
-
-      y = tp->y;
-
-      if(tp->x >= rootWidth / 2) {
-         x = rootWidth - 1;
-      } else {
-         x = 1 - tp->width;
-      }
-
    }
 
-   /* Move it. */
-   JXMoveWindow(display, tp->window, x, y);
+   /* Determine where to move the tray. */
+   switch(layout) {
+   case THIDE_LEFT:
+      x = sp->y - tp->width + 1;
+      y = tp->y;
+      break;
+   case THIDE_RIGHT:
+      x = sp->y + sp->width - 1;
+      y = tp->y;
+      break;
+   case THIDE_TOP:
+      x = tp->x;
+      y = sp->y - tp->height + 1;
+      break;
+   case THIDE_BOTTOM:
+      x = tp->x;
+      y = sp->y + sp->height - 1;
+      break;
+   case THIDE_INVISIBLE:
+      /* Off the top of the screen. */
+      x = tp->x;
+      y = 0 - tp->height - 1;
+      break;
+   default:
+      Assert(0);
+      return;
+   }
 
+   /* Move and redraw. */
+   JXMoveWindow(display, tp->window, x, y);
+   DrawSpecificTray(tp);
 }
 
 /** Process a tray event. */
-int ProcessTrayEvent(const XEvent *event) {
-
+char ProcessTrayEvent(const XEvent *event)
+{
    TrayType *tp;
 
    for(tp = trays; tp; tp = tp->next) {
@@ -573,53 +670,54 @@ int ProcessTrayEvent(const XEvent *event) {
    }
 
    return 0;
-
 }
 
 /** Signal the tray (needed for autohide). */
-void SignalTray(const TimeType *now, int x, int y) {
-
-   TrayType *tp;
-
-   for(tp = trays; tp; tp = tp->next) {
-      if(tp->autoHide && !tp->hidden && !menuShown) {
-         if(x < tp->x || x >= tp->x + tp->width
-            || y < tp->y || y >= tp->y + tp->height) {
-            HideTray(tp);
-         }
-      }
+void SignalTray(const TimeType *now, int x, int y, Window w, void *data)
+{
+   TrayType *tp = (TrayType*)data;
+   Assert(tp->autoHide != THIDE_OFF);
+   if(tp->hidden || menuShown) {
+      return;
    }
 
+   if(x < tp->x || x >= tp->x + tp->width
+      || y < tp->y || y >= tp->y + tp->height) {
+      if(GetTimeDifference(now, &tp->showTime) >= tp->autoHideDelay) {
+         HideTray(tp);
+      }
+   } else {
+      tp->showTime = *now;
+   }
 }
 
 /** Handle a tray expose event. */
-void HandleTrayExpose(TrayType *tp, const XExposeEvent *event) {
-
+void HandleTrayExpose(TrayType *tp, const XExposeEvent *event)
+{
    DrawSpecificTray(tp);
-
 }
 
 /** Handle a tray enter notify (for autohide). */
-void HandleTrayEnterNotify(TrayType *tp, const XCrossingEvent *event) {
-
+void HandleTrayEnterNotify(TrayType *tp, const XCrossingEvent *event)
+{
    ShowTray(tp);
-
 }
 
 /** Get the tray component under the given coordinates. */
-TrayComponentType *GetTrayComponent(TrayType *tp, int x, int y) {
-
+TrayComponentType *GetTrayComponent(TrayType *tp, int x, int y)
+{
    TrayComponentType *cp;
    int xoffset, yoffset;
-   int width, height;
 
-   xoffset = tp->border;
-   yoffset = tp->border;
+   xoffset = 0;
+   yoffset = 0;
    for(cp = tp->components; cp; cp = cp->next) {
-      width = cp->width;
-      height = cp->height;
-      if(x >= xoffset && x < xoffset + width) {
-         if(y >= yoffset && y < yoffset + height) {
+      const int startx = xoffset;
+      const int starty = yoffset;
+      const int width = cp->width;
+      const int height = cp->height;
+      if(x >= startx && x < startx + width) {
+         if(y >= starty && y < starty + height) {
             return cp;
          }
       }
@@ -631,39 +729,32 @@ TrayComponentType *GetTrayComponent(TrayType *tp, int x, int y) {
    }
 
    return NULL;
-
 }
 
 /** Handle a button press on a tray. */
-void HandleTrayButtonPress(TrayType *tp, const XButtonEvent *event) {
-
-   TrayComponentType *cp;
-   int x, y;
-   int mask;
-
-   cp = GetTrayComponent(tp, event->x, event->y);
+void HandleTrayButtonPress(TrayType *tp, const XButtonEvent *event)
+{
+   TrayComponentType *cp = GetTrayComponent(tp, event->x, event->y);
    if(cp && cp->ProcessButtonPress) {
-      x = event->x - cp->x;
-      y = event->y - cp->y;
-      mask = event->button;
+      const int x = event->x - cp->x;
+      const int y = event->y - cp->y;
+      const int mask = event->button;
+      DiscardButtonEvents();
       (cp->ProcessButtonPress)(cp, x, y, mask);
    }
-
 }
 
 /** Handle a button release on a tray. */
-void HandleTrayButtonRelease(TrayType *tp, const XButtonEvent *event) {
-
+void HandleTrayButtonRelease(TrayType *tp, const XButtonEvent *event)
+{
    TrayComponentType *cp;
-   int x, y;
-   int mask;
 
    // First inform any components that have a grab.
    for(cp = tp->components; cp; cp = cp->next) {
       if(cp->grabbed) {
-         x = event->x - cp->x;
-         y = event->y - cp->y;
-         mask = event->button;
+         const int x = event->x - cp->x;
+         const int y = event->y - cp->y;
+         const int mask = event->button;
          (cp->ProcessButtonRelease)(cp, x, y, mask);
          JXUngrabPointer(display, CurrentTime);
          cp->grabbed = 0;
@@ -673,34 +764,28 @@ void HandleTrayButtonRelease(TrayType *tp, const XButtonEvent *event) {
 
    cp = GetTrayComponent(tp, event->x, event->y);
    if(cp && cp->ProcessButtonRelease) {
-      x = event->x - cp->x;
-      y = event->y - cp->y;
-      mask = event->button;
+      const int x = event->x - cp->x;
+      const int y = event->y - cp->y;
+      const int mask = event->button;
       (cp->ProcessButtonRelease)(cp, x, y, mask);
    }
-
 }
 
 /** Handle a motion notify event. */
-void HandleTrayMotionNotify(TrayType *tp, const XMotionEvent *event) {
-
-   TrayComponentType *cp;
-   int x, y;
-   int mask;
-
-   cp = GetTrayComponent(tp, event->x, event->y);
+void HandleTrayMotionNotify(TrayType *tp, const XMotionEvent *event)
+{
+   TrayComponentType *cp = GetTrayComponent(tp, event->x, event->y);
    if(cp && cp->ProcessMotionEvent) {
-      x = event->x - cp->x;
-      y = event->y - cp->y;
-      mask = event->state;
+      const int x = event->x - cp->x;
+      const int y = event->y - cp->y;
+      const int mask = event->state;
       (cp->ProcessMotionEvent)(cp, x, y, mask);
    }
-
 }
 
 /** Draw all trays. */
-void DrawTray() {
-
+void DrawTray(void)
+{
    TrayType *tp;
 
    if(shouldExit) {
@@ -710,85 +795,83 @@ void DrawTray() {
    for(tp = trays; tp; tp = tp->next) {
       DrawSpecificTray(tp);
    }
-
 }
 
 /** Draw a specific tray. */
-void DrawSpecificTray(const TrayType *tp) {
-
+void DrawSpecificTray(const TrayType *tp)
+{
    TrayComponentType *cp;
-   int x;
 
-   Assert(tp);
-
-   /* Draw components. */
    for(cp = tp->components; cp; cp = cp->next) {
       UpdateSpecificTray(tp, cp);
    }
 
-   /* Draw the border. */
-   for(x = 0; x < tp->border; x++) {
-
-      /* Top */
+   if(settings.trayDecorations == DECO_MOTIF) {
       JXSetForeground(display, rootGC, colors[COLOR_TRAY_UP]);
-      JXDrawLine(display, tp->window, rootGC,
-         0, x,
-         tp->width - x - 1, x);
+      JXDrawLine(display, tp->window, rootGC, 0, 0, tp->width - 1, 0);
+      JXDrawLine(display, tp->window, rootGC, 0, tp->height - 1, 0, 0);
 
-      /* Bottom */
       JXSetForeground(display, rootGC, colors[COLOR_TRAY_DOWN]);
-      JXDrawLine(display, tp->window, rootGC,
-         x + 1, tp->height - x - 1,
-         tp->width - x - 2, tp->height - x - 1);
-
-      /* Left */
-      JXSetForeground(display, rootGC, colors[COLOR_TRAY_UP]);
-      JXDrawLine(display, tp->window, rootGC,
-         x, x,
-         x, tp->height - x - 1);
-
-      /* Right */
+      JXDrawLine(display, tp->window, rootGC, 0, tp->height - 1,
+                 tp->width - 1, tp->height - 1);
+      JXDrawLine(display, tp->window, rootGC, tp->width - 1, 0,
+                 tp->width - 1, tp->height - 1);
+   } else {
       JXSetForeground(display, rootGC, colors[COLOR_TRAY_DOWN]);
-      JXDrawLine(display, tp->window, rootGC, 
-         tp->width - x - 1, x + 1,
-         tp->width - x - 1, tp->height - x - 1);
-
+      JXDrawRectangle(display, tp->window, rootGC, 0, 0,
+                      tp->width - 1, tp->height - 1);
    }
+}
 
+/** Raise tray windows. */
+void RaiseTrays(void)
+{
+   TrayType *tp;
+   for(tp = trays; tp; tp = tp->next) {
+      tp->autoHide |= THIDE_RAISED;
+      ShowTray(tp);
+      JXRaiseWindow(display, tp->window);
+   }
+}
+
+/** Lower tray windows. */
+void LowerTrays(void)
+{
+   TrayType *tp;
+   for(tp = trays; tp; tp = tp->next) {
+      tp->autoHide &= ~THIDE_RAISED;
+   }
+   RequireRestack();
 }
 
 /** Update a specific component on a tray. */
-void UpdateSpecificTray(const TrayType *tp, const TrayComponentType *cp) {
-
-   if(cp->pixmap != None && !shouldExit) {
-      JXCopyArea(display, cp->pixmap, tp->window, rootGC, 0, 0,
-         cp->width, cp->height, cp->x, cp->y);
+void UpdateSpecificTray(const TrayType *tp, const TrayComponentType *cp)
+{
+   if(JUNLIKELY(shouldExit)) {
+      return;
    }
 
+   /* If the tray is hidden, draw only the background. */
+   if(cp->pixmap != None) {
+      JXCopyArea(display, cp->pixmap, tp->window, rootGC, 0, 0,
+                 cp->width, cp->height, cp->x, cp->y);
+   }
 }
 
 /** Layout tray components on a tray. */
-void LayoutTray(TrayType *tp, int *variableSize, int *variableRemainder) {
-
+void LayoutTray(TrayType *tp, int *variableSize, int *variableRemainder)
+{
    TrayComponentType *cp;
-   int variableCount;
+   unsigned int variableCount;
    int width, height;
    int temp;
 
-   tp->width = tp->requestedWidth;
-   tp->height = tp->requestedHeight;
-
-   for(cp = tp->components; cp; cp = cp->next) {
-      cp->width = cp->requestedWidth;
-      cp->height = cp->requestedHeight;
-   }
-
-   ComputeTraySize(tp);
+   ComputeTrayGeometry(tp);
 
    /* Get the remaining size after setting fixed size components. */
    /* Also, keep track of the number of variable size components. */
-   width = tp->width - 2 * tp->border;
-   height = tp->height - 2 * tp->border;
+   width = tp->width - TRAY_BORDER_SIZE * 2;
+   height = tp->height - TRAY_BORDER_SIZE * 2;
    variableCount = 0;
    for(cp = tp->components; cp; cp = cp->next) {
       if(tp->layout == LAYOUT_HORIZONTAL) {
@@ -796,14 +879,14 @@ void LayoutTray(TrayType *tp, int *variableSize, int *variableRemainder) {
          if(temp > 0) {
             width -= temp;
          } else {
-            ++variableCount;
+            variableCount += 1;
          }
       } else {
          temp = cp->height;
          if(temp > 0) {
             height -= temp;
          } else {
-            ++variableCount;
+            variableCount += 1;
          }
       }
    }
@@ -834,11 +917,13 @@ void LayoutTray(TrayType *tp, int *variableSize, int *variableRemainder) {
       }
    }
 
+   tp->width = Max(1, tp->width);
+   tp->height = Max(1, tp->height);
 }
 
 /** Resize a tray. */
-void ResizeTray(TrayType *tp) {
-
+void ResizeTray(TrayType *tp)
+{
    TrayComponentType *cp;
    int variableSize;
    int variableRemainder;
@@ -850,8 +935,8 @@ void ResizeTray(TrayType *tp) {
    LayoutTray(tp, &variableSize, &variableRemainder);
 
    /* Reposition items on the tray. */
-   xoffset = tp->border;
-   yoffset = tp->border;
+   xoffset = TRAY_BORDER_SIZE;
+   yoffset = TRAY_BORDER_SIZE;
    for(cp = tp->components; cp; cp = cp->next) {
 
       cp->x = xoffset;
@@ -861,23 +946,23 @@ void ResizeTray(TrayType *tp) {
 
       if(cp->Resize) {
          if(tp->layout == LAYOUT_HORIZONTAL) {
-            height = tp->height - 2 * tp->border;
+            height = tp->height - TRAY_BORDER_SIZE * 2;
             width = cp->width;
             if(width == 0) {
                width = variableSize;
                if(variableRemainder) {
-                  ++width;
-                  --variableRemainder;
+                  width += 1;
+                  variableRemainder -= 1;
                }
             }
          } else {
-            width = tp->width - 2 * tp->border;
+            width = tp->width - TRAY_BORDER_SIZE * 2;
             height = cp->height;
             if(height == 0) {
                height = variableSize;
                if(variableRemainder) {
-                  ++height;
-                  --variableRemainder;
+                  height += 1;
+                  variableRemainder -= 1;
                }
             }
          }
@@ -898,203 +983,154 @@ void ResizeTray(TrayType *tp) {
    }
 
    JXMoveResizeWindow(display, tp->window, tp->x, tp->y,
-      tp->width, tp->height);
+                      tp->width, tp->height);
 
-   UpdateTaskBar();
+   RequireTaskUpdate();
    DrawSpecificTray(tp);
 
    if(tp->hidden) {
       HideTray(tp);
    }
+}
 
+/** Draw the tray background on a drawable. */
+void ClearTrayDrawable(const TrayComponentType *cp)
+{
+   const Drawable d = cp->pixmap != None ? cp->pixmap : cp->window;
+   if(colors[COLOR_TRAY_BG1] == colors[COLOR_TRAY_BG2]) {
+      JXSetForeground(display, rootGC, colors[COLOR_TRAY_BG1]);
+      JXFillRectangle(display, d, rootGC, 0, 0, cp->width, cp->height);
+   } else {
+      DrawHorizontalGradient(d, rootGC, colors[COLOR_TRAY_BG1],
+                             colors[COLOR_TRAY_BG2], 0, 0,
+                             cp->width, cp->height);
+   }
 }
 
 /** Get a linked list of trays. */
-TrayType *GetTrays() {
+TrayType *GetTrays(void)
+{
    return trays;
 }
 
-/** Get a supporting window to use. */
-Window GetSupportingWindow() {
-
-   if(trays) {
-      return trays->window;
-   } else if(supportingWindow != None) {
-      return supportingWindow;
-   } else {
-      supportingWindow = JXCreateSimpleWindow(display, rootWindow,
-         0, 0, 1, 1, 0, 0, 0);
-      return supportingWindow;
-   }
-
+/** Get the number of trays. */
+unsigned int GetTrayCount(void)
+{
+   return trayCount;
 }
 
 /** Determine if a tray should autohide. */
-void SetAutoHideTray(TrayType *tp, int v) {
-   Assert(tp);
-   tp->autoHide = v;
+void SetAutoHideTray(TrayType *tp,
+                     TrayAutoHideType autohide,
+                     unsigned timeout_ms)
+{
+   if(JUNLIKELY(tp->autoHide != THIDE_OFF)) {
+      UnregisterCallback(SignalTray, tp);
+   }
+
+   tp->autoHide = autohide;
+   tp->autoHideDelay = timeout_ms;
+
+   if(autohide != THIDE_OFF) {
+      RegisterCallback(timeout_ms, SignalTray, tp);
+   }
 }
 
 /** Set the x-coordinate of a tray. */
-void SetTrayX(TrayType *tp, const char *str) {
+void SetTrayX(TrayType *tp, const char *str)
+{
    Assert(tp);
    Assert(str);
-   tp->x = atoi(str);
+   tp->requestedX = atoi(str);
 }
 
 /** Set the y-coordinate of a tray. */
-void SetTrayY(TrayType *tp, const char *str) {
+void SetTrayY(TrayType *tp, const char *str)
+{
    Assert(tp);
    Assert(str);
-   tp->y = atoi(str);
+   tp->requestedY = atoi(str);
 }
 
 /** Set the width of a tray. */
-void SetTrayWidth(TrayType *tp, const char *str) {
-
-   int width;
-
-   Assert(tp);
-   Assert(str);
-
-   width = atoi(str);
-
-   if(width < 0) {
-      Warning("invalid tray width: %d", width);
-   } else {
-      tp->requestedWidth = width;
-   }
-
+void SetTrayWidth(TrayType *tp, const char *str)
+{
+   tp->requestedWidth = atoi(str);
 }
 
 /** Set the height of a tray. */
-void SetTrayHeight(TrayType *tp, const char *str) {
-
-   int height;
-
-   Assert(tp);
-   Assert(str);
-
-   height = atoi(str);
-
-   if(height < 0) {
-      Warning("invalid tray height: %d", height);
-   } else {
-      tp->requestedHeight = height;
-   }
+void SetTrayHeight(TrayType *tp, const char *str)
+{
+   tp->requestedHeight = atoi(str);
+}
 
+/** Set the tray screen index. */
+void SetTrayScreen(TrayType *tp, const char *str)
+{
+   tp->screen = atoi(str);
 }
 
 
 /** Set the tray orientation. */
-void SetTrayLayout(TrayType *tp, const char *str) {
-
-   Assert(tp);
-
-   if(!str) {
-
-      /* Compute based on requested size. */
-
-   } else if(!strcmp(str, "horizontal")) {
-
+void SetTrayLayout(TrayType *tp, const char *str)
+{
+   if(!str || !strcmp(str, "horizontal")) {
       tp->layout = LAYOUT_HORIZONTAL;
-      return;
-
    } else if(!strcmp(str, "vertical")) {
-
-      tp->layout = LAYOUT_VERTICAL;
-      return;
-
-   } else {
-      Warning("invalid tray layout: \"%s\"", str);
-   }
-
-   /* Prefer horizontal layout, but use vertical if
-    * width is finite and height is larger than width or infinite.
-    */
-   if(tp->requestedWidth > 0
-      && (tp->requestedHeight == 0
-      || tp->requestedHeight > tp->requestedWidth)) {
       tp->layout = LAYOUT_VERTICAL;
    } else {
-      tp->layout = LAYOUT_HORIZONTAL;
+      Warning(_("invalid tray layout: \"%s\""), str);
    }
-
 }
 
 /** Set the layer for a tray. */
-void SetTrayLayer(TrayType *tp, const char *str) {
-
-   int temp;
-
-   Assert(tp);
-   Assert(str);
-
-   temp = atoi(str);
-   if(temp < LAYER_BOTTOM || temp > LAYER_TOP) {
-      Warning("invalid tray layer: %d", temp);
-      tp->layer = DEFAULT_TRAY_LAYER;
-   } else {
-      tp->layer = temp;
-   }
-
-}
-
-/** Set the border width for a tray. */
-void SetTrayBorder(TrayType *tp, const char *str) {
-
-   int temp;
-
-   Assert(tp);
-   Assert(str);
-
-   temp = atoi(str);
-   if(temp < MIN_TRAY_BORDER || temp > MAX_TRAY_BORDER) {
-      Warning("invalid tray border: %d", temp);
-      tp->border = DEFAULT_TRAY_BORDER;
-   } else {
-      tp->border = temp;
-   }
-
+void SetTrayLayer(TrayType *tp, WinLayerType layer)
+{
+   tp->layer = layer;
 }
 
 /** Set the horizontal tray alignment. */
-void SetTrayHorizontalAlignment(TrayType *tp, const char *str) {
+void SetTrayHorizontalAlignment(TrayType *tp, const char *str)
+{
+   static const StringMappingType mapping[] = {
+      { "center",    TALIGN_CENTER  },
+      { "fixed",     TALIGN_FIXED   },
+      { "left",      TALIGN_LEFT    },
+      { "right",     TALIGN_RIGHT   }
+   };
 
-   Assert(tp);
-
-   if(!str || !strcmp(str, "fixed")) {
+   if(!str) {
       tp->halign = TALIGN_FIXED;
-   } else if(!strcmp(str, "left")) {
-      tp->halign = TALIGN_LEFT;
-   } else if(!strcmp(str, "right")) {
-      tp->halign = TALIGN_RIGHT;
-   } else if(!strcmp(str, "center")) {
-      tp->halign = TALIGN_CENTER;
    } else {
-      Warning("invalid tray horizontal alignment: \"%s\"", str);
-      tp->halign = TALIGN_FIXED;
+      const int x = FindValue(mapping, ARRAY_LENGTH(mapping), str);
+      if(JLIKELY(x >= 0)) {
+         tp->halign = x;
+      } else {
+         Warning(_("invalid tray horizontal alignment: \"%s\""), str);
+         tp->halign = TALIGN_FIXED;
+      }
    }
-
 }
 
 /** Set the vertical tray alignment. */
-void SetTrayVerticalAlignment(TrayType *tp, const char *str) {
+void SetTrayVerticalAlignment(TrayType *tp, const char *str)
+{
+   static const StringMappingType mapping[] = {
+      { "bottom",    TALIGN_BOTTOM  },
+      { "center",    TALIGN_CENTER  },
+      { "fixed",     TALIGN_FIXED   },
+      { "top",       TALIGN_TOP     }
+   };
 
-   Assert(tp);
-
-   if(!str || !strcmp(str, "fixed")) {
+   if(!str) {
       tp->valign = TALIGN_FIXED;
-   } else if(!strcmp(str, "top")) {
-      tp->valign = TALIGN_TOP;
-   } else if(!strcmp(str, "bottom")) {
-      tp->valign = TALIGN_BOTTOM;
-   } else if(!strcmp(str, "center")) {
-      tp->valign = TALIGN_CENTER;
    } else {
-      Warning("invalid tray vertical alignment: \"%s\"", str);
-      tp->valign = TALIGN_FIXED;
+      const int x = FindValue(mapping, ARRAY_LENGTH(mapping), str);
+      if(JLIKELY(x >= 0)) {
+         tp->valign = x;
+      } else {
+         Warning(_("invalid tray vertical alignment: \"%s\""), str);
+         tp->valign = TALIGN_FIXED;
+      }
    }
-
 }
-
